-- Travel window handler

import "Turbine.UI.Lotro"
import "Vinny.Common"
import "Vinny.Common.DropMenu"
import "Vinny.Common.ToolTip"

local labelFont = Turbine.UI.Lotro.Font.TrajanPro14
local foreColor = Turbine.UI.Color( 0.9, 0.9, 0 )
local backColor = Turbine.UI.Color( 0, 0, 0.1 )
local greyColor = Turbine.UI.Color( 0.4, 0.4, 0.4 )
local Alias = Turbine.UI.Lotro.ShortcutType.Alias
local Button = Turbine.UI.Lotro.Button
local CheckBox = Turbine.UI.Lotro.CheckBox
local DropMenu = Vinny.Common.DropMenu
local Item = Turbine.UI.Lotro.ShortcutType.Item
local Label = Turbine.UI.Label
local Skill = Turbine.UI.Lotro.ShortcutType.Skill
local TextBox = Turbine.UI.Lotro.TextBox
local Left = Turbine.UI.ContentAlignment.MiddleLeft
local Center = Turbine.UI.ContentAlignment.MiddleCenter
local Right = Turbine.UI.ContentAlignment.MiddleRight
local Quickslot = Turbine.UI.Lotro.Quickslot
local Qsize = 36
local SetLoc, NoReq, SetEnd, SetFind, Blank
Ms_list = {"1","2","3","4","5","6","7","8","9","10","11"}

function Req(dv)
	if not (TR_req.S0 or dv.S0 or NoReq) then return end
	local code = dv.r
	if not code or NoReq then return true end
	for s in string.gmatch(code,"%u%d+") do
		if not TR_req[s] then return end
	end
	return true
end

local function Action(tbl)
	local code = tbl.r==Race and tbl.rd or tbl.d
	return Turbine.UI.Lotro.Shortcut(Skill,"0x700"..code)
end

local function Find_Loc(name,tbl)
	local loc = tbl[name]
	if loc then
		local aname = loc.n or name
		local act = loc.j and "Journey to" or tbl.nm
		if TR_req[tbl.tb] and TR_req[tbl.tb][aname] then
			print("Found "..act.." "..aname)
     		TR_window.Slot:SetShortcut(Action(loc))
			return true
		end
	end
end

local function Find_Route(Start,End,pl,ss,ht)
	local ix,loc,list,via = 0,'',{Start}, { [Start]={t=-5,c=0,n=0,i=0} }
	while loc do
		ix = ix+1; loc = list[ix]
		if not loc then break end
    	if not Locs[loc] then printe("Nul loc, "..loc) return end
		local v,d,td = via[loc], Locs[loc].d, Locs[loc].td
    	if not d then return end
		td = td and TR_req[td] and 0.9 or 1
		for dest,dv in pairs(d) do
			if not Locs[dest] then printe("Missing dest="..dest.." @ "..loc)
			elseif Locs[dest].d and not TR_req.NV[dest] then
				local l,c,t,s = dv.l, dv.c
				if l and l<0 then l = -l end
				if dv.s and (not l or pl>=l) and Req(dv) then
					s,c,t = true, dv.s, dv.st or 28
				elseif c and (not l or dv.l>0 or pl>=l) then
					t = dv.t or 60
				elseif dv.mt then
					c,t,s = 0,dv.mt,false
				end
				if t then
					t,c = v.t+t+5, v.c+c*td
					if not via[dest] then
						table.insert(list,dest)
						via[dest] = {t=t, c=c, n=v.n+1, s=s, p=loc, i=#list}
					elseif t+ss*c < via[dest].t+ss*via[dest].c then
						vi = via[dest].i
						via[dest] = {t=t, c=c, n=v.n+1, s=s, p=loc, i=vi}
						if ix>vi then ix = vi-1 break end
					end
				end
			end
		end
	end
	local v = via[End]
	if not v then return end
	local V = {n=v.n, c=v.c, t=v.t}
	local str,p,s = End, v.p
	if v.s==false then str = str.."(MT)"
	elseif v.s then str = str.."(ST)" end
	while p do
		v = via[p]
		s = v.s and "(ST)" or v.s==false and "(MT)" or ""
		if v.p or not ht then str = p..s.." -> "..str end
		p = v.p
	end
	V.s = str
  return V
end

-- Find a route between start and end locations
function TR_Route(End,pl,ss)
	local Start = SetLoc
	if Start==End then printe("Select a different End location.") return end
	if SetEnd then Start,End = End,Start end
	printh("Finding a route between "..Start.." and "..End..":")
	local v = Find_Route(Start,End,pl,ss)
	if not v then printe("No usable route found.") return end
	print(string.format("Route in %d steps, Cost=%ss%s",v.n,v.c,TR_Time(v.t)))
	print(v.s)
end

-- Check a Return/Guide/Muster/Sail starting point
function TR_RGM(tbl,skill,pl,ss,V)
  for name,rtt in pairs(skill) do
    local rtn = rtt.n or name
    if tbl[rtn] then
--      print("Checking start at "..rtn)
      local v = Find_Route(name,SetLoc,pl,ss)
      if v then
        local t = v.t+rtt.t+5
        if (not V) or t+ss*v.c <= V.t+ss*V.c then
          V = v
          V.s = skill.nm..' '..rtn.." -> "..v.s
          V.n = V.n+1
          V.a = Action(rtt)
--	print(V.s.." = "..V.t)
        end
      end
    end
  end
  return V
end

-- Find best start location and route to end location
function TR_Start(Start,pl,ss)
	printh("Finding the best start for route to "..SetLoc..":")
	local cnt,V = 0
	if Start and Start~=SetLoc then
		V = Find_Route(Start,SetLoc,pl,ss)
	end
	if not TR_window.skipMs:IsChecked() then
		for ix,name in pairs(TR_req.MS) do
			cnt = cnt+1
			local v = Find_Route(name,SetLoc,pl,ss)
			if v  then
				v.t = v.t + Locs[name].t+10 -- extra penalty for Ms
				if (not V) or v.t+ss*v.c < V.t+ss*V.c then
					V = v
					V.s = "Milestone to "..name.." -> "..v.s
					V.n = V.n+1
					V.a = Turbine.UI.Lotro.Shortcut(Skill,"0x700"..Miles[tonumber(ix)])
				end
--		print(V.s.." = "..V.t)
			end
		end
		if cnt < 1 then
			print("Warning: No milestone locations have been set.")
		end
	end
	local tbl = TR_req.GT or TR_req.MT or TR_req.ST
	if tbl then
		local skill = TR_req.GT and Guide or TR_req.MT and Muster or Sail
    	V = TR_RGM(tbl,skill,pl,ss,V)
	end
	local rt = TR_req.RT
  	if rt and next(rt) then
    	V = TR_RGM(rt,Return,pl,ss,V)
	elseif plevel>30 then
		print("Warning: No Return-to locations have been set.")
	end
	if TR_req.house then
		local hn,ht,v = TR_req.house,TR_req.htime
		local ha = Turbine.UI.Lotro.Shortcut(Skill,"0x700"..House[hn])
		local hs = "Travel to "..hn.."House -> House Stable(MT) -> "
--		print("type(.d)="..type(Locs[Hs].d))
		for name,dt in pairs(Locs[Hs].d) do
			if (not dt.l or pl>= dt.l) and Req(dt) then
				if name==SetLoc then
					v = {t=0, c=dt.s, n=0}
				else v = Find_Route(name,SetLoc,pl,ss,true) end
				if v  then
					v.t = v.t + 12 + ht + dt.st -- add time to get there
					if (not V) or v.t+ss*v.c < V.t+ss*V.c then
						V = v
						V.s = hs..name..((v.s and "(ST) -> "..v.s) or "(ST)")
						V.n = V.n+3
						V.a = ha
					end
				end
			end
		end
		if TR_req.dock then
			local hs = "Travel to "..hn.."House -> House Boat(MT) -> "
			local ht = TR_req.dtime
			for name,dt in pairs(Locs[Dm].d) do
				if (not dt.l or pl>= dt.l) and Req(dt) then
					if name==SetLoc then
						v = {t=0, c=dt.s, n=0}
					else v = Find_Route(name,SetLoc,pl,ss,true) end
					if v  then
						v.t = v.t + 12 + ht + dt.st -- add time to get there
						if (not V) or v.t+ss*v.c < V.t+ss*V.c then
							V = v
							V.s = hs..name..((v.s and "(ST) -> "..v.s) or "(ST)")
							V.n = V.n+3
							V.a = ha
						end
					end
				end
			end
		end
	end
	if not V then printe("No usable route found.") return end
	print(string.format("Route in %d steps, Cost=%ss%s",V.n,V.c,TR_Time(V.t)))
	print(V.s)
	return V.a
end

TR_Window = class( Turbine.UI.Lotro.Window )

function TR_Window:AddField(control, text, pos, size)
	local field = control()
	field:SetParent( self )
	if text then field:SetText( text ) end
	field:SetPosition( pos.x,pos.y )
	field:SetSize( size.x,size.y )
	if control==Button or control==DropMenu or not text then return field end
	field:SetFont( labelFont )
	field:SetBackColor( backColor )
	field:SetForeColor( foreColor )
	field:SetTextAlignment( Center )
	return field
end

---- This is the main Travel Location window
function TR_Window:Constructor()
	Turbine.UI.Lotro.Window.Constructor( self )

	-- Position the window near the top and to left of center of the screen.
	local pos = TR_Opt.pos1 or 
				{ x=Turbine.UI.Display.GetWidth()/5, y=self:GetHeight()/2 }
	self:SetPosition( pos.x, pos.y )
	self:SetText( "Travel Ref." )
	self:SetSize( 250,465 )

	self:AddField(Label, "Level:", {x=20,y=40}, {x=45,y=16} )
	self.level = self:AddField(TextBox, "", {x=60,y=40}, {x=30,y=15} )
	self.level:SetText(plevel)
	Vinny.Common.ToolTip(self.level,-5,-18,"Max Find Route level.",145)

	self:AddField(Label, "Secs/silver:", {x=105,y=40}, {x=80,y=16} )
	self.secs = self:AddField(TextBox, "10", {x=181,y=41}, {x=24,y=15} )
	Vinny.Common.ToolTip(self.secs,-5,-18,"Time vs money route value.",180)
	if TR_req.secs then self.secs:SetText(TR_req.secs) end

	-- Zone label and menu
	self:AddField(Label, "Zone:", {x=20,y=65}, {x=45,y=16} )
	self.zoneMenu = self:AddField(DropMenu, "", {x=90,y=63}, {x=140,y=20} )
	Vinny.Common.ToolTip(self.zoneMenu,2,-20,"Shift for show all.",125)
	local action = function()
		self.locMenu:SetText( "" )
		self.visit:SetForeColor( greyColor )
		self.visit:SetEnabled(false)
		self.visit:SetChecked(false)
	end
	self.zoneMenu.Menu.Click = function() 
		local zlist = {}
		if not self:IsShiftKeyDown() then
			for i,name in ipairs(zones) do
				if plevel>=Zlvl[name] then table.insert(zlist,name) end
			end
		else zlist = zones end
		self.zoneMenu:BuildMenu(zlist,action) 
	end

	-- Location label and menu
	self:AddField(Label, "Location:", {x=20,y=90}, {x=70,y=16} )
	self.locMenu = self:AddField(DropMenu, "", {x=90,y=88}, {x=145,y=20} )
	local action = function(args)
		self.visit:SetEnabled(false)
		self.visit:SetChecked(TR_req.NV[args])
		self.visit:SetForeColor( foreColor )
		self.visit:SetEnabled(true)
	end
	self.locMenu.Menu.Click = function()
        local zone = self.zoneMenu:GetText()
		if zone~="" then
			local Loc_list = {}
			for name,t in pairs(Locs) do
				if t.z==zone and t.d then table.insert(Loc_list,name) end
			end
			table.sort(Loc_list)
			self.locMenu:BuildMenu(Loc_list,action)
		else print("Select a zone.") end
	end

	-- Milestone button and menu
	self.msButton = self:AddField(Button, "Set milestone location", {x=15,y=113}, {x=170,y=20} )
	self.msButton.Click = function( sender,args )
        local loc,ms = self.locMenu:GetText(), self.msMenu:GetText()
        local str = " Cleared."
		if loc ~= "" then
        	if not Locs[loc].t then printe("No milestone defined.") return end
        	local ml =  Locs[loc].ml
        	if ml and plevel<ml then printe("Milestone above player level..") return end
			str = " set to "..loc
        else loc = nil end
        TR_req.MS[ms] = loc
        print("Milestone #"..ms..str)
		Turbine.PluginData.Save(Turbine.DataScope.Character,"Travel_req",TR_req)
	end
	self:AddField(Label, "#", {x=185,y=113}, {x=15,y=20} )
	self.msMenu = self:AddField(DropMenu, "1", {x=197,y=112}, {x=40,y=20} )
	self.msMenu.Menu.Click = function() self.msMenu:BuildMenu(Ms_list) end

	-- Zones button
	self.zonesButton = self:AddField(Button, "List Zones", {x=15,y=140}, {x=90,y=20} )
	Vinny.Common.ToolTip(self.zonesButton,2,-20,"Shift for XP zones.",125)
	self.zonesButton.Click = function( sender,args )
        TR_Command:Execute("tr","zones")
	end

	-- Locations button
	self.locsButton = self:AddField(Button, "List Locations", {x=115,y=139}, {x=120,y=20} )
	self.locsButton.Click = function( sender,args )
        local zone = self.zoneMenu:GetText()
		if zone=="" then print("Select a zone.")
        else TR_Command:Execute("tr",zone) end
	end

	-- Areas button
	self.areasButton = self:AddField(Button, "List Areas", {x=15,y=165}, {x=80,y=20} )
	self.areasButton.Click = function( sender,args )
        TR_Command:Execute("tr","areas")
	end

	-- Destinations button
	self.destButton = self:AddField(Button, "List Destinations", {x=105,y=165}, {x=130,y=20} )
	Vinny.Common.ToolTip(self.destButton,2,-20,"Shift for show all.",125)
	self.destButton.Click = function( sender,args )
        local loc = self.locMenu:GetText()
		if loc~="" then
			local lvl = self.level:GetText()
			if lvl then lvl = tonumber(lvl) end
            if lvl and lvl>0 and lvl<=Max_lvl then
				TR_Command:Execute("tr",loc,lvl,true)
            else print("Error, Invalid level: "..(lvl or "?")) end
        else print("Select a location.") end
	end

	-- Destinations label and menu
	self:AddField(Label, "Dest:", {x=25,y=195}, {x=45,y=16} )
	self.destMenu = self:AddField(DropMenu, "", {x=70,y=193}, {x=145,y=20} )
	local action = function(dest)
        local loc = self.locMenu:GetText()
		TR_Dest( dest, Locs[loc].d[dest], plevel )
		if Locs[dest].n then dest = Locs[dest].n end
		self.zoneMenu:SetText( Locs[dest].z )
		self.locMenu:SetText( dest )
		self.visit:SetChecked(TR_req[dest])
	end
	self.destMenu.Menu.Click = function()
        local loc = self.locMenu:GetText()
		if loc~="" then
			local list = {}
			for name in pairs(Locs[loc].d) do
				if not Locs[name] then
					printe("Missing loc for "..name)
				elseif Locs[name].d or Locs[name].n then
					table.insert(list,name)
				end
			end
			table.sort(list)
			self.destMenu:BuildMenu(list,action)
		else print("Select a location.") end
	end

	-- Start/End button
	self.startButton = self:AddField(Button, "Set Start", {x=15,y=220}, {x=80,y=20} )
	self.startButton.Click = function( sender,args )
        local loc = self.locMenu:GetText()
		if loc~="" then
			SetLoc = loc
			local w =  SetEnd and "End" or "Start"
			print(w.."ing location set to "..SetLoc)
			if SetEnd then
				if Find_Loc(loc,Guide) then return
				elseif Find_Loc(loc,Muster) then return
				elseif Find_Loc(loc,Sail) then return
				elseif Find_Loc(loc,Return) then return end
				for ix,name in pairs(TR_req.MS) do
					if name==loc then
						print("Found milestone to "..name)
						w = Turbine.UI.Lotro.Shortcut(Skill,"0x700"..Miles[tonumber(ix)])
						TR_window.Slot:SetShortcut(w)
						return
					end
				end
			end
			self.Slot:SetShortcut(Blank)
        else print("Select a location.") end
	end

	-- End check box
	local box = self:AddField(CheckBox, "", {x=100,y=222}, {x=25,y=16} )
	Vinny.Common.ToolTip(box,-9,-20,"Start or End button setting.",200)
	if TR_Opt and TR_Opt.SE then 
		SetEnd = true
		self.startButton:SetText( "Set End" )
		box:SetChecked(true)
	end
	box.CheckedChanged = function( sender,args )
		SetEnd = sender:IsChecked()
		self.startButton:SetText( SetEnd and "Set End" or "Set Start" )
		print((SetEnd and "Enabled " or "Disabled ").."Set End.")
	end

	-- Find Route button
	self.routeButton = self:AddField(Button, "Find Route", {x=130,y=220}, {x=100,y=20} )
	self.routeButton.Click = function( sender,args )
		local l = self.level:GetText()
		if l then l = tonumber(l) end
		local ss = self.secs:GetText()
		if ss then ss = tonumber(ss) end
        local loc = self.locMenu:GetText()
		if not SetLoc then print("Set a Start/End location.")
		elseif loc=="" then print("Select a target location.")
		elseif TR_req.NV[loc] then printe(loc.." hasn't been visited.")
		elseif not l or l<1 or l>Max_lvl then printe("Invalid level, "..(l or "?"))
		elseif not ss or ss<0 or ss>99 then printe("Invalid secs.")
        else TR_Route(loc,l,ss) end
	end

	-- No Requirements check box
	local box = self:AddField(CheckBox, "No Req.", {x=30,y=250}, {x=75,y=16} )
	Vinny.Common.ToolTip(box,-2,-20,"Ignore travel requirements.",190)
	box.CheckedChanged = function( sender,args )
		NoReq = sender:IsChecked()
		print((NoReq and "En" or "Dis").."abled No Requirements.")
	end

	-- Not Visit check box
	self.visit = self:AddField(CheckBox, "Not Visited", {x=125,y=250}, {x=95,y=16} )
	self.visit:SetForeColor( greyColor )
	self.visit:SetEnabled(false)
	Vinny.Common.ToolTip(self.visit,-9,-20,"Toggle stable visited flag.",180)
	self.visit.CheckedChanged = function( sender,args )
        local loc = self.locMenu:GetText()
		TR_req.NV[loc] = sender:IsChecked() or nil
		Turbine.PluginData.Save(Turbine.DataScope.Character,"Travel_req",TR_req)
		if sender:IsEnabled() then
			print((sender:IsChecked() and "Disabled " or "Enabled ")..loc)
		end
	end

	-- Find Nearest button
	self.locsButton = self:AddField(Button, "Find nearest Stable", {x=15,y=275}, {x=150,y=20} )
	local slot = Turbine.UI.Lotro.Quickslot()
	slot:SetParent( self.locsButton )
    slot:SetPosition( 2,2 )
    slot:SetSize( 99, 15 )
    slot:SetShortcut(Turbine.UI.Lotro.Shortcut( Alias, "/trl ;loc" ))
    slot:SetAllowDrop( false )

	-- Find check box
	local box = self:AddField(CheckBox, "Rec", {x=175,y=277}, {x=45,y=16} )
	Vinny.Common.ToolTip(box,-9,-20,"Find recruiter stables.",150)
	box.CheckedChanged = function( sender,args )
		SetFind = sender:IsChecked()
		print((SetFind and "En" or "Dis").."abled Find Recruiter")
		slot:SetShortcut(Turbine.UI.Lotro.Shortcut( Alias, SetFind and "/trr ;loc" or "/trl ;loc" ))
	end

	-- Find Best Start button
	self.findButton = self:AddField(Button, "Find best Start", {x=40,y=305}, {x=120,y=20} )
	self.findButton.Click = function( sender,args )
		local l = self.level:GetText()
		if l then l = tonumber(l) end
		local ss = self.secs:GetText()
		if ss then ss = tonumber(ss) end
        local loc = self.locMenu:GetText()
		if not l or l<1 or l>Max_lvl then printe("Invalid level, "..(l or "?"))
		elseif not ss or ss<0 or ss>99 then printe("Invalid secs.")
		elseif TR_req.NV[loc] then printe(loc.." hasn't been visited.")
		elseif SetLoc and SetEnd then
			self.Slot:SetShortcut(TR_Start(loc,l,ss))
		else print("Set an End location.") end
	end

	-- Create a Quickslot
	self.Slot = self:AddField(Quickslot, nil, {x=175,y=305}, {x=Qsize,y=Qsize} )
	self.Slot:SetBackColor( Turbine.UI.Color( 0.2, 0.2, 0.2 ) )
	Blank = self.Slot:GetShortcut()

	-- Skip Milestones check box
	self.skipMs = self:AddField(CheckBox, "Skip Milestones", {x=40,y=330}, {x=125,y=16} )
	Vinny.Common.ToolTip(self.skipMs,-2,-20,"Ignore available milestones.",200)
	self.skipMs.CheckedChanged = function( sender,args )
		print((sender:IsChecked() and "Skipp" or "Us").."ing Milestones.")
	end

	-- Requirements button
	self.reqButton = self:AddField(Button, "Requirements", {x=25,y=355}, {x=105,y=20} )
	self.reqButton.Click = function( sender,args )
        TRW_Command:Execute("trw","tr")
	end

	-- Discounts button
	self.discButton = self:AddField(Button, "Discounts", {x=135,y=355}, {x=85,y=20} )
	self.discButton.Click = function( sender,args )
        TRW_Command:Execute("trw","td")
	end

	-- Return to button
	self.reqButton = self:AddField(Button, "Return", {x=20,y=380}, {x=70,y=20} )
	Vinny.Common.ToolTip(self.reqButton,2,-20,"Shift for learnable skills.",170)
	self.reqButton.Click = function( sender,args )
		if sender:IsShiftKeyDown() then
			TR_Command:Execute("tr","rt")
		else TRW_Command:Execute("trw","rt") end
	end

	-- Guide/Muster/Sail button
	self.gmsButton = self:AddField(Button, "Guide/Muster/Sail", {x=95,y=380}, {x=135,y=20} )
	Vinny.Common.ToolTip(self.gmsButton,-2,-20,"Shift for learnable skills.",170)
	self.gmsButton.Click = function( sender,args )
		if sender:IsShiftKeyDown() then
			TR_Command:Execute("tr","gms")
		else TRW_Command:Execute("trw","gms") end
	end

	-- Find Nearest vault button
	self.findvButton = self:AddField(Button, "Find nearest Vault", {x=20,y=405}, {x=145,y=20} )
	local slot = Turbine.UI.Lotro.Quickslot()
	slot:SetParent( self.findvButton )
    slot:SetPosition( 2,2 )
    slot:SetSize( 99, 15 )
    slot:SetShortcut(Turbine.UI.Lotro.Shortcut( Alias, "/trv ;loc" ))
    slot:SetAllowDrop( false )

	-- Find vaults in zone
	self.vaultButton = self:AddField(Button, "Vaults", {x=170,y=405}, {x=60,y=20} )
	local slot = Turbine.UI.Lotro.Quickslot()
	slot:SetParent( self.vaultButton )
    slot:SetPosition( 2,2 )
    slot:SetSize( 99, 15 )
    slot:SetShortcut(Turbine.UI.Lotro.Shortcut( Alias, "/tr vaults" ))
    slot:SetAllowDrop( false )

	-- Set House Travel button
	self.htButton = self:AddField(Button, "Set House Travel", {x=55,y=430}, {x=140,y=20} )
	self.htButton.Click = function( sender,args )
        TRW_Command:Execute("trw","ht")
	end

end

TR_window = TR_Window()
TR_RWindow = class( Turbine.UI.Lotro.Window )

-- code=req/dest, line=vpos, title=r/g/m, tb=req table, tl=level
function TR_RWindow:AddBox(code,line,title,tb,text)
	local box = Turbine.UI.Lotro.CheckBox()
	local req = tb and TR_req[tb] or TR_req
	local width = self:GetWidth()-50
	box:SetParent( self )
	box:SetPosition( 30,20+20*line )
	box:SetSize( width,20 )
	box:SetBackColor( backColor )
	box:SetTextAlignment( Left )
	box:SetForeColor( foreColor )
	box:SetFont( Turbine.UI.Lotro.Font.TrajanPro16 )
	box:SetText( text )
	box.code = code
	if req[code] then box:SetChecked(true) end
	box.CheckedChanged = function( sender,args )
		local v = sender:IsChecked()
		local str = v and "Enabled " or "Disabled "
		req[code] = v or nil
		print(str..(tb and title.." "..code or Reqs[code]) )
	end
	return box
end

local function code_sort(a,b)
	local a1,b1 = string.sub(a,1,1), string.sub(b,1,1)
	if a1==b1 then return Reqs[a]<Reqs[b]
	else return a1<b1 end
end

function TR_RWindow:Constructor(list,width,title)
	Turbine.UI.Lotro.Window.Constructor( self )
	local skill,tb
	if not title then -- return/guide/muster?
		skill = list
		tb = skill.tb
		if not TR_req[tb] then TR_req[tb] = {} end
		list = {}
		for name,tbl in pairs(skill) do
			if #name>2 then table.insert(list,name) end
		end
		table.sort(list, function(a, b)
			local an = skill[a].n or a
			local bn = skill[b].n or b
			return an<bn
			end)
		title = skill.nm
	end

	-- Position the window near the top and to left of center of the screen.
	pos = { x=Turbine.UI.Display.GetWidth()/3+10, y=self:GetHeight()/3 }
	self:SetPosition( pos.x, pos.y )
 	self:SetSize( width,60+20*#list )
	self:SetText( title )

	-- Create a checkbox and add it to the window.
	l = 0
	for ix,n in ipairs(list) do
		local tbl = skill and skill[n]
		local tl = tbl and tbl.tl
		if not tl or plevel>=tl or (tbl and tbl.r==Race) then -- can it be learned?
			if not tbl or tbl.d or tbl.r==Race then
				local code = tbl and tbl.n or n
				l = l+1
				local text = " "..(tb and code or list==Reqs_list and TR_Req(n) or Reqs[n])
				self:AddBox(code,l,title,tb,text)
			end
		end
	end
 	self:SetSize( width,60+20*l )
end

function TR_RWindow:Closed()
	TR_req.secs = TR_window.secs:GetText()
	Turbine.PluginData.Save(Turbine.DataScope.Character,"Travel_req",TR_req)
	print("Travel settings saved.")
end

-- Home Travel settings
TR_HTWindow = class( Turbine.UI.Lotro.Window )

function TR_HTWindow:AddField(control, text, pos, size)
	local field = control()
	field:SetParent( self )
	if text then field:SetText( text ) end
	field:SetPosition( pos.x,pos.y )
	field:SetSize( size.x,size.y )
	if control==Button or control==DropMenu or not text then return field end
	field:SetFont( labelFont )
	field:SetBackColor( backColor )
	field:SetForeColor( foreColor )
	field:SetTextAlignment( Center )
	return field
end


function TR_HTWindow:Constructor()
	Turbine.UI.Lotro.Window.Constructor( self )

	-- Position the window near the top and to left of center of the screen.
	pos = { x=Turbine.UI.Display.GetWidth()/3+10, y=self:GetHeight()/2 }
	self:SetPosition( pos.x, pos.y )
	self:SetText( "House Travel" )
	self:SetSize( 250,275 )

	-- Create a Quickslot
	self:AddField(Label, "Action:", {x=45,y=150}, {x=55,y=16} )
	self.Slot = self:AddField(Quickslot, nil, {x=110,y=140}, {x=Qsize,y=Qsize} )
	self.Slot:SetBackColor( Turbine.UI.Color( 0.2, 0.2, 0.2 ) )

	-- Home label and menu
	self:AddField(Label, "Select best stable location.", {x=35,y=35}, {x=185,y=16} )
	self:AddField(Label, "House:", {x=20,y=55}, {x=45,y=16} )
	local name,time,dtime,Skiff = None, 20, 20
	if TR_req.house then 
		name = TR_req.house
		time = tostring(TR_req.htime)
		self.Slot:SetShortcut(Turbine.UI.Lotro.Shortcut(Skill,"0x700"..House[name]))
	end
	self.houseMenu = self:AddField(DropMenu, name, {x=70,y=53}, {x=150,y=20} )
	local action = function(args)
		self.saveButton:SetEnabled( true )
		local dest = Blank
		if args~=None then
			dest = Turbine.UI.Lotro.Shortcut(Skill,"0x700"..House[args])
		end
		self.Slot:SetShortcut(dest)
	end
	self.houseMenu.Menu.Click = function() 
		self.houseMenu:BuildMenu(house,action)
	end

	self:AddField(Label, "Stable access time:", {x=20,y=84}, {x=125,y=16} )
	self.time = self:AddField(TextBox, time, {x=147,y=83}, {x=30,y=17} )

	-- Save button
	self.saveButton = self:AddField(Button, "Save settings", {x=20,y=110}, {x=100,y=20} )
	self.saveButton:SetEnabled( false )
	self.saveButton.Click = function( sender,args )
        name = self.houseMenu:GetText()
		if name~=None then
			local time = tonumber(self.time:GetText())
			if not time or time<4 then printe("Invalid time.") return end
			local dtime = tonumber(self.time:GetText())
			if not dtime or dtime<4 then printe("Invalid Dtime.") return end
			TR_req.house = name
			TR_req.htime = time
			TR_req.dock = Skiff or nil
			TR_req.dtime = dtime
		else TR_req.house = nil end
		Turbine.PluginData.Save(Turbine.DataScope.Character,"Travel_req",TR_req)
		print("Travel settings saved.")
	end

	-- List button
	self.listButton = self:AddField(Button, "List Dest.", {x=130,y=110}, {x=100,y=20} )
	self.listButton.Click = function( sender,args )
		TR_Command:Execute("tr",Hs)
	end

	-- Dock-master check box
	box = self:AddField(CheckBox, "Dock-master", {x=45,y=190}, {x=120,y=16} )
	Vinny.Common.ToolTip(box,2,-20,"Dock-master's Skiff.",190)
	box.CheckedChanged = function( sender,args )
		Skiff = sender:IsChecked()
		print((Skiff and "En" or "Dis").."abled Dock-master.")
		self.saveButton:SetEnabled( true )
	end
	box:SetChecked(TR_req.dock)

	self:AddField(Label, "Dock access time:", {x=40,y=211}, {x=125,y=16} )
	self.dtime = self:AddField(TextBox, dtime, {x=167,y=210}, {x=30,y=17} )

	-- List button
	self.dlistButton = self:AddField(Button, "List Docks", {x=70,y=235}, {x=100,y=20} )
	self.dlistButton.Click = function( sender,args )
		TR_Command:Execute("tr",Dm)
	end

end

TR_HTwindow = TR_HTWindow()

if not TR_req.MS then TR_req.MS = {} end
TR_Rwindow = TR_RWindow(Reqs_list,440,"Travel Requirements")
local TDL = {}
for r in pairs(TD_list) do table.insert(TDL,r) end
table.sort(TDL, function(a, b) return Reqs[a]<Reqs[b] end)
TR_Dwindow = TR_RWindow(TDL,400,"Travel Discounts")
TR_RTwindow = TR_RWindow(Return,230)
if PC==162 then TR_GTwindow = TR_RWindow(Guide,230) end
if PC==194 then TR_MTwindow = TR_RWindow(Muster,230) end
if PC==216 then TR_STwindow = TR_RWindow(Sail,230) end

-- Set Escape action
TR_window:SetWantsKeyEvents( true )
TR_window.KeyDown = function(sender, args)
	if( args.Action == Turbine.UI.Lotro.Action.Escape ) then
		TR_window:SetVisible( false )
		TR_Rwindow:SetVisible( false )
		TR_Dwindow:SetVisible( false )
		TR_RTwindow:SetVisible( false )
		TR_HTwindow:SetVisible( false )
	end
end

TRW_Command = Turbine.ShellCommand()
function TRW_Command:GetShortHelp() return "dtw (TR window commands)" end
function TRW_Command:GetHelp() return "trw (open TR window)" end

function TRW_Command:Execute( cmd,args )
    if args=="tr" then TR_Rwindow:SetVisible(true) return end
    if args=="td" then TR_Dwindow:SetVisible(true) return end
    if args=="rt" then TR_RTwindow:SetVisible(true) return end
    if args=="gms" then
    	if PC==194 then TR_MTwindow:SetVisible(true) return end
    	if PC==162 then TR_GTwindow:SetVisible(true) return end
    	if PC==216 then TR_STwindow:SetVisible(true) return end
		printe("Only for Hunters, Wardens or Mariners.")
    end
    if args=="ht" then TR_HTwindow:SetVisible(true) return end
    if args=="?" then Vinny.Common.Help(help,cmd) return end
	if #args==5 and tonumber(args,16) then -- skill id?
		local act = Turbine.UI.Lotro.Shortcut(Skill,"0x700"..args)
		TR_window.Slot:SetShortcut(act)
		return
	end
	if #args>0 then Vinny.Common.Help(help,args) return end
	TR_window.level:SetText( plevel )
    TR_window:SetVisible( true )
end

Turbine.Shell.AddCommand( "trw",TRW_Command )
